'use strict';

var SyntaxJSX = require('@babel/plugin-syntax-jsx');
var t = require('@babel/types');
var helperModuleImports = require('@babel/helper-module-imports');
var htmlEntities = require('html-entities');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () { return e[k]; }
        });
      }
    });
  }
  n["default"] = e;
  return Object.freeze(n);
}

var SyntaxJSX__default = /*#__PURE__*/_interopDefaultLegacy(SyntaxJSX);
var t__namespace = /*#__PURE__*/_interopNamespace(t);

const nonSpreadNameSpaces = new Set([
  "class",
  "style",
  "use",
  "prop",
  "attr"
]);

function getConfig(path) {
  return path.hub.file.metadata.config;
}

const getRendererConfig = (path, renderer) => {
  const config = getConfig(path);
  return config?.renderers?.find(r => r.name === renderer) ?? config;
};

function registerImportMethod(path, name, moduleName) {
  const imports =
    path.scope.getProgramParent().data.imports ||
    (path.scope.getProgramParent().data.imports = new Map());
  moduleName = moduleName || getConfig(path).moduleName;
  if (!imports.has(`${moduleName}:${name}`)) {
    let id = helperModuleImports.addNamed(path, name, moduleName, {
      nameHint: `_$${name}`
    });
    imports.set(`${moduleName}:${name}`, id);
    return id;
  } else {
    let iden = imports.get(`${moduleName}:${name}`);
    // the cloning is required to play well with babel-preset-env which is
    // transpiling import as we add them and using the same identifier causes
    // problems with the multiple identifiers of the same thing
    return t__namespace.cloneDeep(iden);
  }
}

function jsxElementNameToString(node) {
  if (t__namespace.isJSXMemberExpression(node)) {
    return `${jsxElementNameToString(node.object)}.${node.property.name}`;
  }
  if (t__namespace.isJSXIdentifier(node) || t__namespace.isIdentifier(node)) {
    return node.name;
  }
  return `${node.namespace.name}:${node.name.name}`;
}

function getTagName(tag) {
  const jsxName = tag.openingElement.name;
  return jsxElementNameToString(jsxName);
}

function isComponent(tagName) {
  return (
    (tagName[0] && tagName[0].toLowerCase() !== tagName[0]) ||
    tagName.includes(".") ||
    /[^a-zA-Z]/.test(tagName[0])
  );
}

function isDynamic(path, { checkMember, checkTags, checkCallExpressions = true, native }) {
  const config = getConfig(path);
  if (config.generate === "ssr" && native) {
    checkMember = false;
    checkCallExpressions = false;
  }
  const expr = path.node;
  if (t__namespace.isFunction(expr)) return false;
  if (
    expr.leadingComments &&
    expr.leadingComments[0] &&
    expr.leadingComments[0].value.trim() === config.staticMarker
  ) {
    expr.leadingComments.shift();
    return false;
  }
  if (
    (checkCallExpressions && (t__namespace.isCallExpression(expr) || t__namespace.isOptionalCallExpression(expr))) ||
    (checkMember &&
      (t__namespace.isMemberExpression(expr) ||
        t__namespace.isOptionalMemberExpression(expr) ||
        t__namespace.isSpreadElement(expr) ||
        (t__namespace.isBinaryExpression(expr) && expr.operator === "in"))) ||
    (checkTags && (t__namespace.isJSXElement(expr) || t__namespace.isJSXFragment(expr)))
  )
    return true;

  let dynamic;
  path.traverse({
    Function(p) {
      if (t__namespace.isObjectMethod(p.node) && p.node.computed) {
        dynamic = isDynamic(p.get("key"), { checkMember, checkTags, checkCallExpressions, native });
      }
      p.skip();
    },
    CallExpression(p) {
      checkCallExpressions && (dynamic = true) && p.stop();
    },
    OptionalCallExpression(p) {
      checkCallExpressions && (dynamic = true) && p.stop();
    },
    MemberExpression(p) {
      checkMember && (dynamic = true) && p.stop();
    },
    OptionalMemberExpression(p) {
      checkMember && (dynamic = true) && p.stop();
    },
    SpreadElement(p) {
      checkMember && (dynamic = true) && p.stop();
    },
    BinaryExpression(p) {
      checkMember && p.node.operator === "in" && (dynamic = true) && p.stop();
    },
    JSXElement(p) {
      checkTags ? (dynamic = true) && p.stop() : p.skip();
    },
    JSXFragment(p) {
      checkTags ? (dynamic = true) && p.stop() : p.skip();
    }
  });
  return dynamic;
}

function getStaticExpression(path) {
  const node = path.node;
  let value, type;
  return (
    t__namespace.isJSXExpressionContainer(node) &&
    t__namespace.isJSXElement(path.parent) &&
    !isComponent(getTagName(path.parent)) &&
    !t__namespace.isSequenceExpression(node.expression) &&
    (value = path.get("expression").evaluate().value) !== undefined &&
    ((type = typeof value) === "string" || type === "number") &&
    value
  );
}

// remove unnecessary JSX Text nodes
function filterChildren(children) {
  return children.filter(
    ({ node: child }) =>
      !(t__namespace.isJSXExpressionContainer(child) && t__namespace.isJSXEmptyExpression(child.expression)) &&
      (!t__namespace.isJSXText(child) || !/^[\r\n]\s*$/.test(child.extra.raw))
  );
}

function checkLength(children) {
  let i = 0;
  children.forEach(path => {
    const child = path.node;
    !(t__namespace.isJSXExpressionContainer(child) && t__namespace.isJSXEmptyExpression(child.expression)) &&
      (!t__namespace.isJSXText(child) || !/^\s*$/.test(child.extra.raw)) &&
      i++;
  });
  return i > 1;
}

function trimWhitespace(text) {
  text = text.replace(/\r/g, "");
  if (/\n/g.test(text)) {
    text = text
      .split("\n")
      .map((t, i) => (i ? t.replace(/^\s*/g, "") : t))
      .filter(s => !/^\s*$/.test(s))
      .join(" ");
  }
  return text.replace(/\s+/g, " ");
}

function transformCondition$1(path, inline, deep) {
  const config = getConfig(path);
  const expr = path.node;
  const memo = registerImportMethod(path, config.memoWrapper);
  let dTest, cond, id;
  if (
    t__namespace.isConditionalExpression(expr) &&
    (isDynamic(path.get("consequent"), {
      checkTags: true
    }) ||
      isDynamic(path.get("alternate"), { checkTags: true }))
  ) {
    dTest = isDynamic(path.get("test"), { checkMember: true });
    if (dTest) {
      cond = expr.test;
      if (!t__namespace.isBinaryExpression(cond))
        cond = t__namespace.unaryExpression("!", t__namespace.unaryExpression("!", cond, true), true);
      id = inline
        ? t__namespace.callExpression(memo, [t__namespace.arrowFunctionExpression([], cond)])
        : path.scope.generateUidIdentifier("_c$");
      expr.test = t__namespace.callExpression(id, []);
      if (t__namespace.isConditionalExpression(expr.consequent) || t__namespace.isLogicalExpression(expr.consequent)) {
        expr.consequent = transformCondition$1(path.get("consequent"), inline, true);
      }
      if (t__namespace.isConditionalExpression(expr.alternate) || t__namespace.isLogicalExpression(expr.alternate)) {
        expr.alternate = transformCondition$1(path.get("alternate"), inline, true);
      }
    }
  } else if (t__namespace.isLogicalExpression(expr)) {
    let nextPath = path;
    // handle top-level or, ie cond && <A/> || <B/>
    while (nextPath.node.operator !== "&&" && t__namespace.isLogicalExpression(nextPath.node.left)) {
      nextPath = nextPath.get("left");
    }
    nextPath.node.operator === "&&" &&
      isDynamic(nextPath.get("right"), { checkTags: true }) &&
      (dTest = isDynamic(nextPath.get("left"), {
        checkMember: true
      }));
    if (dTest) {
      cond = nextPath.node.left;
      if (!t__namespace.isBinaryExpression(cond))
        cond = t__namespace.unaryExpression("!", t__namespace.unaryExpression("!", cond, true), true);
      id = inline
        ? t__namespace.callExpression(memo, [t__namespace.arrowFunctionExpression([], cond)])
        : path.scope.generateUidIdentifier("_c$");
      nextPath.node.left = t__namespace.callExpression(id, []);
    }
  }
  if (dTest && !inline) {
    const statements = [
      t__namespace.variableDeclaration("const", [
        t__namespace.variableDeclarator(
          id,
          config.memoWrapper
            ? t__namespace.callExpression(memo, [t__namespace.arrowFunctionExpression([], cond)])
            : t__namespace.arrowFunctionExpression([], cond)
        )
      ]),
      t__namespace.arrowFunctionExpression([], expr)
    ];
    return deep
      ? t__namespace.callExpression(
          t__namespace.arrowFunctionExpression(
            [],
            t__namespace.blockStatement([statements[0], t__namespace.returnStatement(statements[1])])
          ),
          []
        )
      : statements;
  }
  return deep ? expr : t__namespace.arrowFunctionExpression([], expr);
}

function escapeBackticks(value) {
  return value.replace(/`/g, "\\`");
}

function escapeHTML(s, attr) {
  if (typeof s !== "string") return s;
  const delim = attr ? '"' : "<";
  const escDelim = attr ? "&quot;" : "&lt;";
  let iDelim = s.indexOf(delim);
  let iAmp = s.indexOf("&");

  if (iDelim < 0 && iAmp < 0) return s;

  let left = 0,
    out = "";

  while (iDelim >= 0 && iAmp >= 0) {
    if (iDelim < iAmp) {
      if (left < iDelim) out += s.substring(left, iDelim);
      out += escDelim;
      left = iDelim + 1;
      iDelim = s.indexOf(delim, left);
    } else {
      if (left < iAmp) out += s.substring(left, iAmp);
      out += "&amp;";
      left = iAmp + 1;
      iAmp = s.indexOf("&", left);
    }
  }

  if (iDelim >= 0) {
    do {
      if (left < iDelim) out += s.substring(left, iDelim);
      out += escDelim;
      left = iDelim + 1;
      iDelim = s.indexOf(delim, left);
    } while (iDelim >= 0);
  } else {
    while (iAmp >= 0) {
      if (left < iAmp) out += s.substring(left, iAmp);
      out += "&amp;";
      left = iAmp + 1;
      iAmp = s.indexOf("&", left);
    }
  }

  return left < s.length ? out + s.substring(left) : out;
}

function convertJSXIdentifier(node) {
  if (t__namespace.isJSXIdentifier(node)) {
    if (t__namespace.isValidIdentifier(node.name)) {
      node.type = "Identifier";
    } else {
      return t__namespace.stringLiteral(node.name);
    }
  } else if (t__namespace.isJSXMemberExpression(node)) {
    return t__namespace.memberExpression(
      convertJSXIdentifier(node.object),
      convertJSXIdentifier(node.property)
    );
  } else if (t__namespace.isJSXNamespacedName(node)) {
    return t__namespace.stringLiteral(`${node.namespace.name}:${node.name.name}`);
  }

  return node;
}

function canNativeSpread(key, { checkNameSpaces } = {}) {
  if (checkNameSpaces && key.includes(":") && nonSpreadNameSpaces.has(key.split(":")[0])) return false;
  // TODO: figure out how to detect definitely function ref
  if (key === "ref") return false;
  return true;
}

function transformElement$1(path, info) {
  let tagName = getTagName(path.node),
    results = {
      id: path.scope.generateUidIdentifier("el$"),
      decl: [],
      exprs: [],
      dynamics: [],
      postExprs: [],
      tagName,
      renderer: "universal"
    };

  results.decl.push(
    t__namespace.variableDeclarator(
      results.id,
      t__namespace.callExpression(
        registerImportMethod(
          path,
          "createElement",
          getRendererConfig(path, "universal").moduleName
        ),
        [t__namespace.stringLiteral(tagName)]
      )
    )
  );

  transformAttributes(path, results);
  transformChildren(path, results);

  return results;
}

function transformAttributes(path, results) {
  let children, spreadExpr;
  let attributes = path.get("openingElement").get("attributes");
  const elem = results.id,
    hasChildren = path.node.children.length > 0,
    config = getConfig(path);

  // preprocess spreads
  if (attributes.some(attribute => t__namespace.isJSXSpreadAttribute(attribute.node))) {
    [attributes, spreadExpr] = processSpreads(path, attributes, {
      elem,
      hasChildren,
      wrapConditionals: config.wrapConditionals
    });
    path.get("openingElement").set(
      "attributes",
      attributes.map(a => a.node)
    );
  }

  path
    .get("openingElement")
    .get("attributes")
    .forEach(attribute => {
      const node = attribute.node;

      let value = node.value,
        key = t__namespace.isJSXNamespacedName(node.name)
          ? `${node.name.namespace.name}:${node.name.name.name}`
          : node.name.name,
        reservedNameSpace = t__namespace.isJSXNamespacedName(node.name) && node.name.namespace.name === "use";
      if (
        t__namespace.isJSXNamespacedName(node.name) &&
        reservedNameSpace &&
        !t__namespace.isJSXExpressionContainer(value)
      ) {
        node.value = value = t__namespace.JSXExpressionContainer(value || t__namespace.JSXEmptyExpression());
      }
      if (t__namespace.isJSXExpressionContainer(value)) {
        if (key === "ref") {
          // Normalize expressions for non-null and type-as
          while (
            t__namespace.isTSNonNullExpression(value.expression) ||
            t__namespace.isTSAsExpression(value.expression)
          ) {
            value.expression = value.expression.expression;
          }
          if (t__namespace.isLVal(value.expression)) {
            const refIdentifier = path.scope.generateUidIdentifier("_ref$");
            results.exprs.unshift(
              t__namespace.variableDeclaration("const", [
                t__namespace.variableDeclarator(refIdentifier, value.expression)
              ]),
              t__namespace.expressionStatement(
                t__namespace.conditionalExpression(
                  t__namespace.binaryExpression(
                    "===",
                    t__namespace.unaryExpression("typeof", refIdentifier),
                    t__namespace.stringLiteral("function")
                  ),
                  t__namespace.callExpression(
                    registerImportMethod(
                      path,
                      "use",
                      getRendererConfig(path, "universal").moduleName
                    ),
                    [refIdentifier, elem]
                  ),
                  t__namespace.assignmentExpression("=", value.expression, elem)
                )
              )
            );
          } else if (t__namespace.isFunction(value.expression)) {
            results.exprs.unshift(
              t__namespace.expressionStatement(
                t__namespace.callExpression(
                  registerImportMethod(
                    path,
                    "use",
                    getRendererConfig(path, "universal").moduleName
                  ),
                  [value.expression, elem]
                )
              )
            );
          } else if (t__namespace.isCallExpression(value.expression)) {
            const refIdentifier = path.scope.generateUidIdentifier("_ref$");
            results.exprs.unshift(
              t__namespace.variableDeclaration("const", [
                t__namespace.variableDeclarator(refIdentifier, value.expression)
              ]),
              t__namespace.expressionStatement(
                t__namespace.logicalExpression(
                  "&&",
                  t__namespace.binaryExpression(
                    "===",
                    t__namespace.unaryExpression("typeof", refIdentifier),
                    t__namespace.stringLiteral("function")
                  ),
                  t__namespace.callExpression(
                    registerImportMethod(
                      path,
                      "use",
                      getRendererConfig(path, "universal").moduleName
                    ),
                    [refIdentifier, elem]
                  )
                )
              )
            );
          }
        } else if (key.startsWith("use:")) {
          // Some trick to treat JSXIdentifier as Identifier
          node.name.name.type = "Identifier";
          results.exprs.unshift(
            t__namespace.expressionStatement(
              t__namespace.callExpression(
                registerImportMethod(path, "use", getRendererConfig(path, "universal").moduleName),
                [
                  node.name.name,
                  elem,
                  t__namespace.arrowFunctionExpression(
                    [],
                    t__namespace.isJSXEmptyExpression(value.expression)
                      ? t__namespace.booleanLiteral(true)
                      : value.expression
                  )
                ]
              )
            )
          );
        } else if (key === "children") {
          children = value;
        } else if (
          config.effectWrapper &&
          isDynamic(attribute.get("value").get("expression"), {
            checkMember: true
          })
        ) {
          results.dynamics.push({ elem, key, value: value.expression });
        } else {
          results.exprs.push(
            t__namespace.expressionStatement(setAttr(attribute, elem, key, value.expression))
          );
        }
      } else {
        results.exprs.push(t__namespace.expressionStatement(setAttr(attribute, elem, key, value)));
      }
    });
  if (spreadExpr) results.exprs.push(spreadExpr);
  if (!hasChildren && children) {
    path.node.children.push(children);
  }
}

function setAttr(path, elem, name, value, { prevId } = {}) {
  if (!value) value = t__namespace.booleanLiteral(true);
  return t__namespace.callExpression(
    registerImportMethod(path, "setProp", getRendererConfig(path, "universal").moduleName),
    prevId ? [elem, t__namespace.stringLiteral(name), value, prevId] : [elem, t__namespace.stringLiteral(name), value]
  );
}

function transformChildren(path, results) {
  const filteredChildren = filterChildren(path.get("children")),
    multi = checkLength(filteredChildren),
    childNodes = filteredChildren.map(transformNode).reduce((memo, child) => {
      if (!child) return memo;
      const i = memo.length;
      if (child.text && i && memo[i - 1].text) {
        memo[i - 1].template += child.template;
      } else memo.push(child);
      return memo;
    }, []);

  const appends = [];
  childNodes.forEach((child, index) => {
    if (!child) return;
    if (child.tagName && child.renderer !== "universal") {
      throw new Error(`<${child.tagName}> is not supported in <${getTagName(path.node)}>.
        Wrap the usage with a component that would render this element, eg. Canvas`);
    }
    if (child.id) {
      let insertNode = registerImportMethod(
        path,
        "insertNode",
        getRendererConfig(path, "universal").moduleName
      );
      let insert = child.id;
      if (child.text) {
        let createTextNode = registerImportMethod(
          path,
          "createTextNode",
          getRendererConfig(path, "universal").moduleName
        );
        if (multi) {
          results.decl.push(
            t__namespace.variableDeclarator(
              child.id,
              t__namespace.callExpression(createTextNode, [
                t__namespace.templateLiteral([t__namespace.templateElement({ raw: child.template })], [])
              ])
            )
          );
        } else
          insert = t__namespace.callExpression(createTextNode, [
            t__namespace.templateLiteral([t__namespace.templateElement({ raw: child.template })], [])
          ]);
      }
      appends.push(t__namespace.expressionStatement(t__namespace.callExpression(insertNode, [results.id, insert])));
      results.decl.push(...child.decl);
      results.exprs.push(...child.exprs);
      results.dynamics.push(...child.dynamics);
    } else if (child.exprs.length) {
      let insert = registerImportMethod(
        path,
        "insert",
        getRendererConfig(path, "universal").moduleName
      );
      if (multi) {
        results.exprs.push(
          t__namespace.expressionStatement(
            t__namespace.callExpression(insert, [
              results.id,
              child.exprs[0],
              nextChild(childNodes, index) || t__namespace.nullLiteral()
            ])
          )
        );
      } else {
        results.exprs.push(
          t__namespace.expressionStatement(t__namespace.callExpression(insert, [results.id, child.exprs[0]]))
        );
      }
    }
  });
  results.exprs.unshift(...appends);
}

function nextChild(children, index) {
  return children[index + 1] && (children[index + 1].id || nextChild(children, index + 1));
}

function processSpreads(path, attributes, { elem, hasChildren, wrapConditionals }) {
  // TODO: skip but collect the names of any properties after the last spread to not overwrite them
  const filteredAttributes = [];
  const spreadArgs = [];
  let runningObject = [];
  let dynamicSpread = false;
  let firstSpread = false;
  attributes.forEach(attribute => {
    const node = attribute.node;
    const key =
      !t__namespace.isJSXSpreadAttribute(node) &&
      (t__namespace.isJSXNamespacedName(node.name)
        ? `${node.name.namespace.name}:${node.name.name.name}`
        : node.name.name);
    if (t__namespace.isJSXSpreadAttribute(node)) {
      firstSpread = true;
      if (runningObject.length) {
        spreadArgs.push(t__namespace.objectExpression(runningObject));
        runningObject = [];
      }
      spreadArgs.push(
        isDynamic(attribute.get("argument"), {
          checkMember: true
        }) && (dynamicSpread = true)
          ? t__namespace.isCallExpression(node.argument) &&
            !node.argument.arguments.length &&
            !t__namespace.isCallExpression(node.argument.callee) &&
            !t__namespace.isMemberExpression(node.argument.callee)
            ? node.argument.callee
            : t__namespace.arrowFunctionExpression([], node.argument)
          : node.argument
      );
    } else if (
      (firstSpread ||
        (t__namespace.isJSXExpressionContainer(node.value) &&
          isDynamic(attribute.get("value").get("expression"), { checkMember: true }))) &&
      canNativeSpread(key)
    ) {
      const isContainer = t__namespace.isJSXExpressionContainer(node.value);
      const dynamic =
        isContainer && isDynamic(attribute.get("value").get("expression"), { checkMember: true });
      if (dynamic) {
        const id = convertJSXIdentifier(node.name);
        let expr =
          wrapConditionals &&
          (t__namespace.isLogicalExpression(node.value.expression) || t__namespace.isConditionalExpression(node.value.expression))
            ? transformCondition(attribute.get("value").get("expression"), true)
            : t__namespace.arrowFunctionExpression([], node.value.expression);
        runningObject.push(
          t__namespace.objectMethod(
            "get",
            id,
            [],
            t__namespace.blockStatement([t__namespace.returnStatement(expr.body)]),
            !t__namespace.isValidIdentifier(key)
          )
        );
      } else {
        runningObject.push(
          t__namespace.objectProperty(
            t__namespace.stringLiteral(key),
            isContainer ? node.value.expression : node.value || t__namespace.stringLiteral("")
          )
        );
      }
    } else filteredAttributes.push(attribute);
  });

  if (runningObject.length) {
    spreadArgs.push(t__namespace.objectExpression(runningObject));
  }

  const props =
    spreadArgs.length === 1 && !dynamicSpread
      ? spreadArgs[0]
      : t__namespace.callExpression(registerImportMethod(path, "mergeProps"), spreadArgs);

  return [
    filteredAttributes,
    t__namespace.expressionStatement(
      t__namespace.callExpression(
        registerImportMethod(path, "spread", getRendererConfig(path, "universal").moduleName),
        [elem, props, t__namespace.booleanLiteral(hasChildren)]
      )
    )
  ];
}

function createTemplate(path, result, wrap) {
  const config = getConfig(path);
  if (result.id) {
    result.decl = t__namespace.variableDeclaration("const", result.decl);
    if (
      !(result.exprs.length || result.dynamics.length || result.postExprs.length) &&
      result.decl.declarations.length === 1
    ) {
      return result.decl.declarations[0].init;
    } else {
      return t__namespace.callExpression(
        t__namespace.arrowFunctionExpression(
          [],
          t__namespace.blockStatement([
            result.decl,
            ...result.exprs.concat(
              wrapDynamics(path, result.dynamics) || [],
              result.postExprs || []
            ),
            t__namespace.returnStatement(result.id)
          ])
        ),
        []
      );
    }
  }
  if (wrap && result.dynamic && config.memoWrapper) {
    return t__namespace.callExpression(registerImportMethod(path, config.memoWrapper), [result.exprs[0]]);
  }
  return result.exprs[0];
}

function wrapDynamics(path, dynamics) {
  if (!dynamics.length) return;
  const config = getConfig(path);

  const effectWrapperId = registerImportMethod(path, config.effectWrapper);

  if (dynamics.length === 1) {
    const prevValue = t__namespace.identifier("_$p");

    return t__namespace.expressionStatement(
      t__namespace.callExpression(effectWrapperId, [
        t__namespace.arrowFunctionExpression(
          [prevValue],
          setAttr(path, dynamics[0].elem, dynamics[0].key, dynamics[0].value, {
            dynamic: true,
            prevId: prevValue
          })
        )
      ])
    );
  }
  const decls = [],
    statements = [],
    identifiers = [],
    prevId = t__namespace.identifier("_p$");
  dynamics.forEach(({ elem, key, value }) => {
    const identifier = path.scope.generateUidIdentifier("v$");
    identifiers.push(identifier);
    decls.push(t__namespace.variableDeclarator(identifier, value));
    const prev = t__namespace.memberExpression(prevId, identifier);
    statements.push(
      t__namespace.expressionStatement(
        t__namespace.logicalExpression(
          "&&",
          t__namespace.binaryExpression("!==", identifier, t__namespace.memberExpression(prevId, identifier)),
          t__namespace.assignmentExpression("=", t__namespace.memberExpression(prevId, identifier), setAttr(
            path,
            elem,
            key,
            identifier,
            { dynamic: true, prevId: prev }
          ))
        )
      )
    );
  });

  return t__namespace.expressionStatement(
    t__namespace.callExpression(effectWrapperId, [
      t__namespace.arrowFunctionExpression(
        [prevId],
        t__namespace.blockStatement([
          t__namespace.variableDeclaration("const", decls),
          ...statements,
          t__namespace.returnStatement(prevId)
        ])
      ),
      t__namespace.objectExpression(identifiers.map(id => t__namespace.objectProperty(id, t__namespace.identifier("undefined"))))
    ])
  );
}

function convertComponentIdentifier(node) {
  if (t__namespace.isJSXIdentifier(node)) {
    if (t__namespace.isValidIdentifier(node.name)) node.type = "Identifier";
    else return t__namespace.stringLiteral(node.name);
  } else if (t__namespace.isJSXMemberExpression(node)) {
    const prop = convertComponentIdentifier(node.property);
    const computed = t__namespace.isStringLiteral(prop);
    return t__namespace.memberExpression(convertComponentIdentifier(node.object), prop, computed);
  }

  return node;
}

function transformComponent(path) {
  let exprs = [],
    config = getConfig(path),
    tagId = convertComponentIdentifier(path.node.openingElement.name),
    props = [],
    runningObject = [],
    dynamicSpread = false,
    hasChildren = path.node.children.length > 0;

  if (config.builtIns.indexOf(tagId.name) > -1 && !path.scope.hasBinding(tagId.name)) {
    const newTagId = registerImportMethod(path, tagId.name);
    tagId.name = newTagId.name;
  }

  path
    .get("openingElement")
    .get("attributes")
    .forEach(attribute => {
      const node = attribute.node;
      if (t__namespace.isJSXSpreadAttribute(node)) {
        if (runningObject.length) {
          props.push(t__namespace.objectExpression(runningObject));
          runningObject = [];
        }
        props.push(
          isDynamic(attribute.get("argument"), {
            checkMember: true
          }) && (dynamicSpread = true)
            ? t__namespace.isCallExpression(node.argument) &&
              !node.argument.arguments.length &&
              !t__namespace.isCallExpression(node.argument.callee) &&
              !t__namespace.isMemberExpression(node.argument.callee)
              ? node.argument.callee
              : t__namespace.arrowFunctionExpression([], node.argument)
            : node.argument
        );
      } else {
        const value = node.value || t__namespace.booleanLiteral(true),
          id = convertJSXIdentifier(node.name),
          key = id.name;
        if (hasChildren && key === "children") return;
        if (t__namespace.isJSXExpressionContainer(value))
          if (key === "ref") {
            if (config.generate === "ssr") return;
            // Normalize expressions for non-null and type-as
            while (
              t__namespace.isTSNonNullExpression(value.expression) ||
              t__namespace.isTSAsExpression(value.expression)
            ) {
              value.expression = value.expression.expression;
            }
            let binding,
              isFunction =
                t__namespace.isIdentifier(value.expression) &&
                (binding = path.scope.getBinding(value.expression.name)) &&
                binding.kind === "const";
            if (!isFunction && t__namespace.isLVal(value.expression)) {
              const refIdentifier = path.scope.generateUidIdentifier("_ref$");
              runningObject.push(
                t__namespace.objectMethod(
                  "method",
                  t__namespace.identifier("ref"),
                  [t__namespace.identifier("r$")],
                  t__namespace.blockStatement([
                    t__namespace.variableDeclaration("const", [
                      t__namespace.variableDeclarator(refIdentifier, value.expression)
                    ]),
                    t__namespace.expressionStatement(
                      t__namespace.conditionalExpression(
                        t__namespace.binaryExpression(
                          "===",
                          t__namespace.unaryExpression("typeof", refIdentifier),
                          t__namespace.stringLiteral("function")
                        ),
                        t__namespace.callExpression(refIdentifier, [t__namespace.identifier("r$")]),
                        t__namespace.assignmentExpression("=", value.expression, t__namespace.identifier("r$"))
                      )
                    )
                  ])
                )
              );
            } else if (isFunction || t__namespace.isFunction(value.expression)) {
              runningObject.push(t__namespace.objectProperty(t__namespace.identifier("ref"), value.expression));
            } else if (t__namespace.isCallExpression(value.expression)) {
              const refIdentifier = path.scope.generateUidIdentifier("_ref$");
              runningObject.push(
                t__namespace.objectMethod(
                  "method",
                  t__namespace.identifier("ref"),
                  [t__namespace.identifier("r$")],
                  t__namespace.blockStatement([
                    t__namespace.variableDeclaration("const", [
                      t__namespace.variableDeclarator(refIdentifier, value.expression)
                    ]),
                    t__namespace.expressionStatement(
                      t__namespace.logicalExpression(
                        "&&",
                        t__namespace.binaryExpression(
                          "===",
                          t__namespace.unaryExpression("typeof", refIdentifier),
                          t__namespace.stringLiteral("function")
                        ),
                        t__namespace.callExpression(refIdentifier, [t__namespace.identifier("r$")])
                      )
                    )
                  ])
                )
              );
            }
          } else if (
            isDynamic(attribute.get("value").get("expression"), {
              checkMember: true,
              checkTags: true
            })
          ) {
            let expr =
              config.wrapConditionals &&
              config.generate !== "ssr" &&
              (t__namespace.isLogicalExpression(value.expression) ||
                t__namespace.isConditionalExpression(value.expression))
                ? transformCondition$1(attribute.get("value").get("expression"), true)
                : t__namespace.arrowFunctionExpression([], value.expression);
            runningObject.push(
              t__namespace.objectMethod(
                "get",
                id,
                [],
                t__namespace.blockStatement([t__namespace.returnStatement(expr.body)]),
                !t__namespace.isValidIdentifier(key)
              )
            );
          } else runningObject.push(t__namespace.objectProperty(id, value.expression));
        else runningObject.push(t__namespace.objectProperty(id, value));
      }
    });

  const childResult = transformComponentChildren(path.get("children"), config);
  if (childResult && childResult[0]) {
    if (childResult[1]) {
      const body =
        t__namespace.isCallExpression(childResult[0]) && t__namespace.isFunction(childResult[0].callee)
          ? childResult[0].callee.body
          : childResult[0].body;
      runningObject.push(
        t__namespace.objectMethod(
          "get",
          t__namespace.identifier("children"),
          [],
          t__namespace.isExpression(body) ? t__namespace.blockStatement([t__namespace.returnStatement(body)]) : body
        )
      );
    } else runningObject.push(t__namespace.objectProperty(t__namespace.identifier("children"), childResult[0]));
  }
  if (runningObject.length || !props.length) props.push(t__namespace.objectExpression(runningObject));

  if (props.length > 1 || dynamicSpread) {
    props = [t__namespace.callExpression(registerImportMethod(path, "mergeProps"), props)];
  }
  const componentArgs = [tagId, props[0]];
  exprs.push(t__namespace.callExpression(registerImportMethod(path, "createComponent"), componentArgs));

  // handle hoisting conditionals
  if (exprs.length > 1) {
    const ret = exprs.pop();
    exprs = [
      t__namespace.callExpression(
        t__namespace.arrowFunctionExpression([], t__namespace.blockStatement([...exprs, t__namespace.returnStatement(ret)])),
        []
      )
    ];
  }
  return { exprs, template: "", component: true };
}

function transformComponentChildren(children, config) {
  const filteredChildren = filterChildren(children);
  if (!filteredChildren.length) return;
  let dynamic = false;

  let transformedChildren = filteredChildren.reduce((memo, path) => {
    if (t__namespace.isJSXText(path.node)) {
      const v = htmlEntities.decode(trimWhitespace(path.node.extra.raw));
      if (v.length) memo.push(t__namespace.stringLiteral(v));
    } else {
      const child = transformNode(path, {
        topLevel: true,
        componentChild: true
      });
      dynamic = dynamic || child.dynamic;
      if (
        config.generate === "ssr" &&
        filteredChildren.length > 1 &&
        child.dynamic &&
        t__namespace.isFunction(child.exprs[0])
      ) {
        child.exprs[0] = child.exprs[0].body;
      }
      memo.push(getCreateTemplate()(path, child, filteredChildren.length > 1));
    }
    return memo;
  }, []);

  if (transformedChildren.length === 1) {
    transformedChildren = transformedChildren[0];
    if (
      !t__namespace.isJSXExpressionContainer(filteredChildren[0]) &&
      !t__namespace.isJSXSpreadChild(filteredChildren[0]) &&
      !t__namespace.isJSXText(filteredChildren[0])
    ) {
      transformedChildren =
        t__namespace.isCallExpression(transformedChildren) &&
        !transformedChildren.arguments.length &&
        !t__namespace.isIdentifier(transformedChildren.callee)
          ? transformedChildren.callee
          : t__namespace.arrowFunctionExpression([], transformedChildren);
      dynamic = true;
    }
  } else {
    transformedChildren = t__namespace.arrowFunctionExpression([], t__namespace.arrayExpression(transformedChildren));
    dynamic = true;
  }
  return [transformedChildren, dynamic];
}

function transformFragmentChildren(children, results, config) {
  const filteredChildren = filterChildren(children),
    childNodes = filteredChildren.reduce((memo, path) => {
      if (t__namespace.isJSXText(path.node)) {
        const v = htmlEntities.decode(trimWhitespace(path.node.extra.raw));
        if (v.length) memo.push(t__namespace.stringLiteral(v));
      } else {
        const child = transformNode(path, { topLevel: true, fragmentChild: true });
        memo.push(getCreateTemplate()(path, child, true));
      }
      return memo;
    }, []);
  results.exprs.push(childNodes.length === 1 ? childNodes[0] : t__namespace.arrayExpression(childNodes));
}

function transformJSX(path) {
    getConfig(path);
    const replace = transformThis(path);
    const result = transformNode(
        path,
        t__namespace.isJSXFragment(path.node)
            ? {}
            : {
                  topLevel: true
              }
    );

    const template = getCreateTemplate();

    path.replaceWith(replace(template(path, result, false)));
}

function transformThis(path) {
    let thisId;
    path.traverse({
        ThisExpression(path) {
            thisId || (thisId = path.scope.generateUidIdentifier('self$'));
            path.replaceWith(thisId);
        }
    });
    return node => {
        if (thisId) {
            let parent = path.getStatementParent();
            const decl = t__namespace.variableDeclaration('const', [
                t__namespace.variableDeclarator(thisId, t__namespace.thisExpression())
            ]);
            parent.insertBefore(decl);
        }
        return node;
    };
}

function transformNode(path, info = {}) {
    const config = getConfig(path);
    const node = path.node;
    let staticValue;
    if (t__namespace.isJSXElement(node)) {
        return transformElement(config, path, info);
    } else if (t__namespace.isJSXFragment(node)) {
        let results = { template: '', decl: [], exprs: [], dynamics: [] };
        // <><div /><Component /></>
        transformFragmentChildren(path.get('children'), results);
        return results;
    } else if (t__namespace.isJSXText(node) || (staticValue = getStaticExpression(path))) {
        const text =
            staticValue !== undefined
                ? !info.doNotEscape
                    ? escapeHTML(staticValue.toString())
                    : staticValue.toString()
                : trimWhitespace(node.extra.raw);
        if (!text.length) return null;
        const results = {
            template: escapeBackticks(text),
            decl: [],
            exprs: [],
            dynamics: [],
            postExprs: [],
            text: true
        };
        if (!info.skipId) results.id = path.scope.generateUidIdentifier('el$');
        return results;
    } else if (t__namespace.isJSXExpressionContainer(node)) {
        if (t__namespace.isJSXEmptyExpression(node.expression)) return null;
        if (
            !isDynamic(path.get('expression'), {
                checkMember: true,
                checkTags: !!info.componentChild,
                native: !info.componentChild
            })
        ) {
            return { exprs: [node.expression], template: '' };
        }
        const expr =
            config.wrapConditionals &&
            config.generate !== 'ssr' &&
            (t__namespace.isLogicalExpression(node.expression) ||
                t__namespace.isConditionalExpression(node.expression))
                ? transformCondition$1(
                      path.get('expression'),
                      info.componentChild
                  )
                : !info.componentChild &&
                  (config.generate !== 'ssr' || info.fragmentChild) &&
                  t__namespace.isCallExpression(node.expression) &&
                  !t__namespace.isMemberExpression(node.expression.callee) &&
                  node.expression.arguments.length === 0
                ? node.expression.callee
                : t__namespace.arrowFunctionExpression([], node.expression);
        return {
            exprs:
                expr.length > 1
                    ? [
                          t__namespace.callExpression(
                              t__namespace.arrowFunctionExpression(
                                  [],
                                  t__namespace.blockStatement([
                                      expr[0],
                                      t__namespace.returnStatement(expr[1])
                                  ])
                              ),
                              []
                          )
                      ]
                    : [expr],
            template: '',
            dynamic: true
        };
    } else if (t__namespace.isJSXSpreadChild(node)) {
        if (
            !isDynamic(path.get('expression'), {
                checkMember: true,
                native: !info.componentChild
            })
        )
            return { exprs: [node.expression], template: '' };
        const expr = t__namespace.arrowFunctionExpression([], node.expression);
        return {
            exprs: [expr],
            template: '',
            dynamic: true
        };
    }
}

function getCreateTemplate(config, path, result) {
    return createTemplate;
}

function transformElement(config, path, info = {}) {
    const node = path.node;
    let tagName = getTagName(node);
    // <Component ...></Component>
    if (isComponent(tagName)) return transformComponent(path);
    for (var renderer of config.renderers ?? []) {
        if (renderer.elements.indexOf(tagName) !== -1) {
            break;
        }
    }

    return transformElement$1(path);
}

// add to the top/bottom of the module.
var postprocess = path => {};

var config = {
  moduleName: "dom",
  generate: "dom",
  hydratable: false,
  delegateEvents: true,
  delegatedEvents: [],
  builtIns: [],
  requireImportSource: false,
  wrapConditionals: true,
  contextToCustomElements: false,
  staticMarker: "@once",
  effectWrapper: "effect",
  memoWrapper: "memo"
};

var preprocess = (path, { opts })  => {
  const merged = path.hub.file.metadata.config = Object.assign({}, config, opts);
  const lib = merged.requireImportSource;
  if (lib) {
    const comments = path.hub.file.ast.comments;
    for(let i = 0; i < comments.length; i++) {
      const comment = comments[i];
      const index = comment.value.indexOf("@jsxImportSource");
      if (index > -1 && comment.value.slice(index).includes(lib)) return;
    }
    path.skip();
  }
};

var index = () => {
  return {
    name: "JSX DOM Expressions",
    inherits: SyntaxJSX__default["default"],
    visitor: {
      JSXElement: transformJSX,
      JSXFragment: transformJSX,
      Program: {
        enter: preprocess,
        exit: postprocess
      }
    }
  };
};

module.exports = index;
